// Port.cpp: implementation of the CPort class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Serial758UI.h"
#include "Port.h"
#include "Status.h"
#include "xrapi.h"
#include "WinIoCtl.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPort::CPort()
{

}

CPort::~CPort()
{

}

HANDLE CPort::CreatePort(unsigned char cPortNo)
{
	//this function opens the specified port and sets the baud and other parameters and
	//also sets the flow control and timeouts. The argument cPortNo represents the port
	//to be opened.This function creates only one port at each call
	
	CSerial758UIApp* pApp=(CSerial758UIApp*)AfxGetApp();
	
	CString ComName;
	COMMTIMEOUTS to;
	DCB dcb;
	HANDLE PortHandle;
	
	long int nBaudarray[14]={110,300,600,1200,2400,4800,9600,19200,38400,57600,115200,230400,460800,921600};
	int nDataBitsarray[4]={5,6,7,8};
	char cParityarray[5]={'O','E','M','S','N'};
	BYTE dwParityarray[5]={ODDPARITY,EVENPARITY,MARKPARITY,SPACEPARITY,NOPARITY};
	BYTE dwStopBitsarray[3]={ ONESTOPBIT, ONE5STOPBITS, TWOSTOPBITS };
	
	int nxrDTRRTSHys[5] = {0,4,6,8,16};//uncomment when used with the exar card
	int nxrTxTrig[65],nxrRxTrig[65];
	
	FillMemory(&dcb,sizeof(dcb),0);//initialize all the values to zero to start with
	dcb.DCBlength=sizeof(dcb);
	int nIndex=cPortNo-(pApp->m_structPortSettingsDlg[0].nComBase + 1);
	
	m_hPortHandle = PortHandle = NULL;//initialize to NULL
	if (cPortNo <= 9)
		ComName.Format("COM%d",cPortNo);//portno has any value from 1 to 9
	else
		ComName.Format("\\\\.\\COM%d",cPortNo);//portno has any value greater than 9
	if((PortHandle=CreateFile(ComName,GENERIC_READ|GENERIC_WRITE,0,0,OPEN_EXISTING,FILE_FLAG_OVERLAPPED,0)) == INVALID_HANDLE_VALUE)  //NON OVERLAPPED READ AND WRITE
	{
		DWORD err=GetLastError();
		return (NULL);//if port creation failed
	}
	else//port was successfully opened
	{	
		m_hPortHandle = PortHandle;
		if(!SetupComm(PortHandle,4096,4096))//set the input buffer size
			AfxGetMainWnd()->MessageBox("Error while setting up the Rx and Tx buffer sizes","ERROR",MB_OK|MB_ICONEXCLAMATION);
		
		if(!GetCommState(PortHandle,&dcb))//get the current DCB settings
			AfxGetMainWnd()->MessageBox("Error while getting the current DCB settings","ERROR",MB_OK|MB_ICONEXCLAMATION);
		
		// update DCB rate, byte size, parity, and stop bits size
		
		dcb.BaudRate = nBaudarray[pApp->m_structPortSettingsDlg[nIndex].nBaud];
		dcb.ByteSize = nDataBitsarray[pApp->m_structPortSettingsDlg[nIndex].nDataBits];
		dcb.Parity   = dwParityarray[pApp->m_structPortSettingsDlg[nIndex].nParity];
		dcb.StopBits = dwStopBitsarray[pApp->m_structPortSettingsDlg[nIndex].nStopBits];
		
		//set the required handshake parameters
		dcb.fDtrControl=DTR_CONTROL_ENABLE ;//this is common for all handshaking methods
		
		//enable the hardware handshaking(CTS/RTS)
		if(!pApp->m_structPortSettingsDlg[nIndex].nFlowCtrl)
		{
			dcb.fOutxCtsFlow=TRUE;
			dcb.fOutX=FALSE;
			dcb.fInX=FALSE;
			dcb.fOutxDsrFlow=TRUE;
			dcb.fRtsControl=RTS_CONTROL_HANDSHAKE;
		}
		//enable the sotfware handshaking(XON/XOFF)
		else if(pApp->m_structPortSettingsDlg[nIndex].nFlowCtrl==1)
		{
			dcb.fOutX=TRUE;
			dcb.fInX=TRUE;
			dcb.fTXContinueOnXoff=TRUE;	
			dcb.fOutxCtsFlow=FALSE;
			dcb.fOutxDsrFlow=FALSE;
			dcb.fRtsControl=RTS_CONTROL_ENABLE;
		}
		//None-no handshaking used
		else 
		{
			dcb.fOutxCtsFlow=FALSE;
			dcb.fOutX=FALSE;
			dcb.fInX=FALSE;
			dcb.fOutxDsrFlow=FALSE;
			dcb.fRtsControl=RTS_CONTROL_ENABLE;
		}
		
		dcb.fBinary=TRUE;//required for proper operation				
		//update the timeouts
		
		memset(&to,0,sizeof(to));
		//to.ReadIntervalTimeout = 0xFFFFFFFF;
		to.ReadTotalTimeoutMultiplier  = 1; //25;
		to.ReadTotalTimeoutConstant =100; //1000;//before it was 500
		//	to.ReadTotalTimeoutMultiplier =0;
		//to.ReadTotalTimeoutConstant =1000;
		to.WriteTotalTimeoutMultiplier  = 1; //10;
		to.WriteTotalTimeoutConstant =100; //1000;		
		
		PurgeComm(PortHandle,PURGE_TXCLEAR|PURGE_RXCLEAR|PURGE_TXABORT|PURGE_RXABORT);//PURGE_RXCLEAR
		
		SetCommTimeouts(PortHandle,&to);//set the new timeouts
		SetCommState(PortHandle,&dcb);//set the new dcb  properties
		
		//the following code should be unmasked when using exar ports
		for(int p=0;p<64;p++)
		{
			nxrTxTrig[p]=p;
			nxrRxTrig[p]=p+1;
		}
		nxrRxTrig[63]=63;
		COM_PORT_ENHANCED	portEnhanced;
    	DWORD  cbReturned; 
		
		if(!pApp->m_structPortSettingsDlg[nIndex].nFlowCtrl)//hardware handshaking
			portEnhanced.FlowControl = 0xF1;
		else if(pApp->m_structPortSettingsDlg[nIndex].nFlowCtrl == 1)//software handshaking
			portEnhanced.FlowControl = 0xF0;
		else
			portEnhanced.FlowControl = 0xF3;  //no handshaking  
		portEnhanced.TxFifoTriggerLevel = nxrTxTrig[pApp->m_structPortSettingsDlg[nIndex].nTxTrig];
		portEnhanced.RxFifoTriggerLevel = nxrRxTrig[pApp->m_structPortSettingsDlg[nIndex].nRxTrig];
		if(pApp->m_structPortSettingsDlg[nIndex].nDTR)
			portEnhanced.Hysteresis = nxrDTRRTSHys[pApp->m_structPortSettingsDlg[nIndex].nDTRHys];
		else
            portEnhanced.Hysteresis = nxrDTRRTSHys[pApp->m_structPortSettingsDlg[nIndex].nRTSHys];
			
		portEnhanced.LoopBack = 0x0;
		portEnhanced.RS485 = (pApp->m_structPortSettingsDlg[nIndex].nRS485) ? 0x20 : 0x00;
		portEnhanced.IR = (pApp->m_structPortSettingsDlg[nIndex].nIR) ? 0x20 : 0x00;
		if (!DeviceIoControl( m_hPortHandle,
			(DWORD) IOCTL_XRPORT_SET_ENHANCED,
			&portEnhanced,
			sizeof(COM_PORT_ENHANCED),
			NULL, 
			0, 
			&cbReturned,
			0
			))
		{
			MessageBox( NULL, "Unable to pass data into driver", NULL, MB_OK);
		}
		return (m_hPortHandle);//return the port handle
	}
}

bool CPort::WritePort(unsigned char* WriteBuff, int WriteBytes)
{
	//this function writes to the port from WriteBuff.Writing is done asynchronously.

	CSerial758UIApp* pApp=(CSerial758UIApp*)AfxGetApp();
	//m_hPortHandle=pApp->m_structPort.hPortHandle[0];

	DWORD nActualWrite;
	CString strWrite,dwdelay;
	strWrite.Format("Unable to Write to the comm port"); 
	
	OVERLAPPED osWrite = {0};  // initialize the overlapped structure for write operations
	    
    osWrite.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);// create this writes overlapped structure hEvent
	if(osWrite.hEvent == NULL)
	{
		MessageBox(AfxGetMainWnd()->m_hWnd,"Error creating overlapped write event handle","Write Error",MB_OK);
		return FALSE;
	}

	int bResult=WriteFile(m_hPortHandle,WriteBuff,WriteBytes,&nActualWrite,&osWrite);//overlapped write operation

	if(!bResult)//if write function fails
	{
		if(GetLastError() == ERROR_IO_PENDING)// asynchronous i/o is still in progress ,do something else for a while             
		{

#ifndef OLD_INFINITE_CHECKING 
#define TIMEOUT	10000 // 10 secs
			DWORD dwRes;
			
			dwRes = WaitForSingleObject(osWrite.hEvent, TIMEOUT);
			
			if (WAIT_OBJECT_0 == dwRes)
			{
				if(GetOverlappedResult(m_hPortHandle, &osWrite, &nActualWrite, FALSE) )
				{
					//
					// Write is completed successfully
					//					
					CloseHandle(osWrite.hEvent);
					return TRUE;						
				}
			//	else
			//	{
			//		MessageBox(AfxGetMainWnd()->m_hWnd,"Write Timeout","Write Error",MB_OK);
			//		CloseHandle(osWrite.hEvent);
			//		return FALSE;
			//	}
			}
			//else
			//{
			//
			// if it comes here it has to be error
			//
				MessageBox(AfxGetMainWnd()->m_hWnd,"Write Timeout","Write Error",MB_OK);
				CloseHandle(osWrite.hEvent);
				return FALSE;

			//}
			
			
#else
				DWORD endTime;
			endtime=GetTickCount() + 10000;
			while(!GetOverlappedResult(m_hPortHandle, &osWrite, &nActualWrite, FALSE))
			{
				if(GetTickCount() > endtime)//time lasped and the write was not complete
				{
					MessageBox(AfxGetMainWnd()->m_hWnd,"Write Timeout","Write Error",MB_OK);
					CloseHandle(osWrite.hEvent);
					return FALSE;
				}				
			}
			CloseHandle(osWrite.hEvent);
			return TRUE;
#endif
		}//end of if error_io_pending
		else//if any error other than error_io_pending
		{
			DWORD dwerr=GetLastError();
		   	MessageBox(AfxGetMainWnd()->m_hWnd,strWrite,"Write Error",MB_OK);		
			CloseHandle(osWrite.hEvent);
			return FALSE;
		}
	}
	else//if the function was successfull and completed immediately
	{
		CloseHandle(osWrite.hEvent);
		return TRUE;	
	}
}

bool CPort::ReadPort(unsigned char IBuff[1000],int *nbytes)
{
	//this function reads the com port and copies the read bytes into
	//IBuff and the no.of bytes into bytes

	CSerial758UIApp* pApp=(CSerial758UIApp*)AfxGetApp();

	DWORD dwErrorMask;
	DWORD nToRead;
	COMSTAT comstat;
	int bResult;
	CString strRead;
	strRead.Format("Unable to Read to the comm port"); 
	OVERLAPPED osRead = {0};  // overlapped structure for write operations
	
    osRead.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);//create this read overlapped structure hEvent
	if(osRead.hEvent == NULL)
	{	
		MessageBox(AfxGetMainWnd()->m_hWnd,"Error creating the overlapped read event handle","Read Error",MB_OK);
		return FALSE;					
	}

	ClearCommError(m_hPortHandle,&dwErrorMask,&comstat);//to set the comstat variables
	nToRead=comstat.cbInQue;//get the no. of bytes waiting in the que
	if(nToRead>450)
		nToRead=450;//read 450 bytes at a time
	
	if(nToRead>0)
	{
		bResult=ReadFile(m_hPortHandle,IBuff,nToRead,&nToRead,&osRead);//overlapped functionality not used.just dummy
		
		if(!bResult)//if write function fails
		{
			if(GetLastError() == ERROR_IO_PENDING)// asynchronous i/o is still in progress ,do something else for a while             
			{
#ifndef OLD_INFINITE_CHECKING 
#define TIMEOUT	10000 // 10 secs
			DWORD dwRes;
			
			dwRes = WaitForSingleObject(osRead.hEvent, TIMEOUT);
			
			if (WAIT_OBJECT_0 == dwRes)
			{
				if(GetOverlappedResult(m_hPortHandle, &osRead, &nToRead, FALSE))
				{
					//
					// read is completed successfully
					//
					*nbytes=(int)nToRead;
					CloseHandle(osRead.hEvent);
					return TRUE;						
				}
			//	else
			//	{
			//		MessageBox(AfxGetMainWnd()->m_hWnd,"Write Timeout","Write Error",MB_OK);
			//		CloseHandle(osWrite.hEvent);
			//		return FALSE;
			//	}
			}
			//else
			//{
			//
			// if it comes here it has to be error
			//
				MessageBox(AfxGetMainWnd()->m_hWnd,strRead,"Read Error",MB_OK);
				CloseHandle(osRead.hEvent);
				return FALSE;

			//}
			
			
#else
				DWORD endTime;
				endtime=GetTickCount() + 10000;
				while(!GetOverlappedResult(m_hPortHandle, &osRead, &nToRead, FALSE))
				{
					if(GetTickCount() > endtime)//time lasped and the write was not complete
					{
						MessageBox(AfxGetMainWnd()->m_hWnd,"Read Timeout","Read Error",MB_OK);
						CloseHandle(osRead.hEvent);
						//PurgeComm(m_hPortHandle,PURGE_RXCLEAR);//clear the extra bytes
						return FALSE;
					}
				}				
				
				*nbytes=(int)nToRead;
				CloseHandle(osRead.hEvent);
				//PurgeComm(m_hPortHandle,PURGE_RXCLEAR);//clear the extra bytes
				return TRUE;
#endif							
			}//end of if error_io_pending
			else//if any error other than error_io_pending
			{
				MessageBox(AfxGetMainWnd()->m_hWnd,strRead,"Read Error",MB_OK);
				CloseHandle(osRead.hEvent);
				//PurgeComm(m_hPortHandle,PURGE_RXCLEAR);//clear the extra bytes
				return FALSE;
			}
		}
		else//if the function was successfull
		{
			*nbytes=(int)nToRead;
			CloseHandle(osRead.hEvent);
			//PurgeComm(m_hPortHandle,PURGE_RXCLEAR);//clear the extra bytes
			return TRUE;	
		}
	}//if nToRead==0
	else
	{
		CloseHandle(osRead.hEvent);
		//PurgeComm(m_hPortHandle,PURGE_RXCLEAR);//clear the extra bytes
		return FALSE;//if no. of bytes read is  0	
	}
}

void CPort::SetHandle(HANDLE hPortHandle)
{
	//sets the handle to its data member
	m_hPortHandle = hPortHandle;

}
